/* QLogic (R)NIC Driver/Library
 * Copyright (c) 2015-2016  QLogic Corporation
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef _QED_ROCE_H
#define _QED_ROCE_H
#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include "qed.h"
#include "qed_dev_api.h"
#include "qed_hsi.h"
#include "qed_if.h"
#include "qed_roce_if.h"

struct qed_rdma_resize_cq_in_params {
	/* input variables (given by miniport) */

	u16 icid;
	u32 cq_size;
	bool pbl_two_level;
	u64 pbl_ptr;
	u16 pbl_num_pages;
	u8 pbl_page_size_log;	/* for the pages that contain the
				 * pointers to the CQ pages
				 */
};

struct qed_rdma_resize_cq_out_params {
	/* output variables, provided to the upper layer */
	u32 prod;		/* CQ producer value on old PBL */
	u32 cons;		/* CQ consumer value on old PBL */
};

struct qed_rdma_resize_cnq_in_params {
	/* input variables (given by miniport) */
	u32 cnq_id;
	u32 pbl_page_size_log;	/* for the pages that contain the
				 * pointers to the cnq pages
				 */
	u64 pbl_ptr;
};

int
qed_rdma_add_user(void *rdma_cxt,
		  struct qed_rdma_add_user_out_params *out_params);

int qed_rdma_alloc_pd(void *rdma_cxt, u16 * pd);

int qed_rdma_alloc_tid(void *rdma_cxt, u32 * tid);

int
qed_rdma_create_cq(void *rdma_cxt,
		   struct qed_rdma_create_cq_in_params *params, u16 * icid);

/* Returns a pointer to the responders' CID, which is also a pointer to the
 * qed_qp_params struct. Returns NULL in case of failure.
 */
struct qed_rdma_qp *qed_rdma_create_qp(void *rdma_cxt,
				       struct qed_rdma_create_qp_in_params
				       *in_params,
				       struct qed_rdma_create_qp_out_params
				       *out_params);

int
qed_roce_create_ud_qp(void *rdma_cxt,
		      struct qed_rdma_create_qp_out_params *out_params);

int qed_rdma_deregister_tid(void *rdma_cxt, u32 tid);

int
qed_rdma_destroy_cq(void *rdma_cxt,
		    struct qed_rdma_destroy_cq_in_params *in_params,
		    struct qed_rdma_destroy_cq_out_params *out_params);

int qed_rdma_destroy_qp(void *rdma_cxt, struct qed_rdma_qp *qp);

int qed_roce_destroy_ud_qp(void *rdma_cxt, u16 cid);

void qed_rdma_free_pd(void *rdma_cxt, u16 pd);

void qed_rdma_free_tid(void *rdma_cxt, u32 tid);

int
qed_rdma_modify_qp(void *rdma_cxt,
		   struct qed_rdma_qp *qp,
		   struct qed_rdma_modify_qp_in_params *params);

struct qed_rdma_device *qed_rdma_query_device(void *rdma_cxt);

struct qed_rdma_port *qed_rdma_query_port(void *rdma_cxt);

int
qed_rdma_query_qp(void *rdma_cxt,
		  struct qed_rdma_qp *qp,
		  struct qed_rdma_query_qp_out_params *out_params);

int
qed_rdma_register_tid(void *rdma_cxt,
		      struct qed_rdma_register_tid_in_params *params);

void qed_rdma_remove_user(void *rdma_cxt, u16 dpi);

int
qed_rdma_resize_cnq(void *rdma_cxt,
		    struct qed_rdma_resize_cnq_in_params *in_params);

/*Returns the CQ CID or zero in case of failure */
int
qed_rdma_resize_cq(void *rdma_cxt,
		   struct qed_rdma_resize_cq_in_params *in_params,
		   struct qed_rdma_resize_cq_out_params *out_params);

/* Before calling rdma_start upper layer (VBD/qed) should fill the
 * page-size and mtu in hwfn context
 */
int qed_rdma_start(void *p_hwfn, struct qed_rdma_start_in_params *params);

int qed_rdma_stop(void *rdma_cxt);

int
qed_rdma_query_stats(void *rdma_cxt,
		     u8 stats_queue,
		     struct qed_rdma_stats_out_params *out_parms);

int
qed_rdma_query_counters(void *rdma_cxt,
			struct qed_rdma_counters_out_params *out_parms);

u32 qed_rdma_get_sb_id(void *p_hwfn, u32 rel_sb_id);

u32 qed_rdma_query_cau_timer_res(void *p_hwfn);

void qed_rdma_cnq_prod_update(void *rdma_cxt, u8 cnq_index, u16 prod);

void qed_rdma_resc_free(struct qed_hwfn *p_hwfn);

#ifdef CONFIG_IWARP

/* iWARP API */

int
qed_iwarp_connect(void *rdma_cxt,
		  struct qed_iwarp_connect_in *iparams,
		  struct qed_iwarp_connect_out *oparams);

int
qed_iwarp_create_listen(void *rdma_cxt,
			struct qed_iwarp_listen_in *iparams,
			struct qed_iwarp_listen_out *oparams);

int qed_iwarp_accept(void *rdma_cxt, struct qed_iwarp_accept_in *iparams);

int qed_iwarp_reject(void *rdma_cxt, struct qed_iwarp_reject_in *iparams);

int qed_iwarp_destroy_listen(void *rdma_cxt, void *handle);

int qed_iwarp_send_rtr(void *rdma_cxt, struct qed_iwarp_send_rtr_in *iparams);

int qed_iwarp_tcp_abort(void *rdma_cxt, struct qed_iwarp_tcp_abort_in *iparams);

#endif /* CONFIG_IWARP */

/* Constants */

/* HW/FW RoCE Limitations (internal. For external see qed_rdma_api.h) */
#define QED_RDMA_MAX_FMR                    (RDMA_MAX_TIDS)	/* 2^17 - 1 */
#define QED_RDMA_MAX_P_KEY                  (1)
#define QED_RDMA_MAX_WQE                    (0x7FFF)	/* 2^15 -1 */
#define QED_RDMA_MAX_SRQ_WQE_ELEM           (0x7FFF)	/* 2^15 -1 */
#define QED_RDMA_PAGE_SIZE_CAPS             (0xFFFFF000)	/* TODO: > 4k?! */
#define QED_RDMA_ACK_DELAY                  (15)	/* 131 milliseconds */
#define QED_RDMA_MAX_MR_SIZE                (0x10000000000ULL)	/* 2^40 */
#define QED_RDMA_MAX_CQS                    (RDMA_MAX_CQS)	/* 64k */
#define QED_RDMA_MAX_MRS                    (RDMA_MAX_TIDS)	/* 2^17 - 1 */
/* Add 1 for header element */
#define QED_RDMA_MAX_SRQ_ELEM_PER_WQE         (RDMA_MAX_SGE_PER_RQ_WQE + 1)
#define QED_RDMA_MAX_SGE_PER_SRQ_WQE          (RDMA_MAX_SGE_PER_RQ_WQE)
#define QED_RDMA_SRQ_WQE_ELEM_SIZE          (16)
#define QED_RDMA_MAX_SRQS                     (32 * 1024)	/* 32k */

/* Configurable */
/* Max CQE is derived from u16/32 size, halved and decremented by 1 to handle
 * wrap properly and then decremented by 1 again. The latter decrement comes
 * from a requirement to create a chain that is bigger than what the user
 * requested by one:
 * The CQE size is 32 bytes but the FW writes in chunks of 64
 * bytes, for performance purposes. Allocating an extra entry and telling the
 * FW we have less prevents overwriting the first entry in case of a wrap i.e.
 * when the FW writes the last entry and the application hasn't read the first
 * one.
 */
#define QED_RDMA_MAX_CQE_32_BIT             (0x7FFFFFFF - 1)
#define QED_RDMA_MAX_CQE_16_BIT             (0x7FFF - 1)

enum qed_rdma_toggle_bit {
	QED_RDMA_TOGGLE_BIT_CLEAR = 0,
	QED_RDMA_TOGGLE_BIT_SET = 1
};

/* @@@TBD Currently we support only affilited events
 * enum qed_rdma_unaffiliated_event_code {
 * QED_RDMA_PORT_ACTIVE, // Link Up
 * QED_RDMA_PORT_CHANGED, // SGID table has changed
 * QED_RDMA_LOCAL_CATASTROPHIC_ERR, // Fatal device error
 * QED_RDMA_PORT_ERR, // Link down
 * };
 */

#define QEDR_MAX_BMAP_NAME      (10)
struct qed_bmap {
	u32 max_count;
	unsigned long *bitmap;
	char name[QEDR_MAX_BMAP_NAME];
};

/* functions for enabling/disabling edpm in rdma PFs according to existence of
 * qps during DCBx update or bar size
 */
void qed_roce_dpm_dcbx(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt);
void qed_rdma_dpm_bar(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt);

#ifdef CONFIG_IWARP

#define QED_IWARP_LL2_SYN_TX_SIZE       (128)
#define QED_IWARP_LL2_SYN_RX_SIZE       (256)

#define QED_IWARP_LL2_OOO_DEF_TX_SIZE   (256)
#define QED_IWARP_LL2_OOO_DEF_RX_SIZE   (4096)

#define QED_IWARP_MAX_SYN_PKT_SIZE      (128)
#define QED_IWARP_HANDLE_INVAL  (0xff)

struct qed_iwarp_ll2_buff {
	struct qed_iwarp_ll2_buff *piggy_buf;
	void *data;
	dma_addr_t data_phys_addr;
	u32 buff_size;
};

struct qed_iwarp_ll2_mpa_buf {
	struct list_head list_entry;
	struct qed_iwarp_ll2_buff *ll2_buf;
	struct unaligned_opaque_data data;
	u16 tcp_payload_len;
	u8 placement_offset;
};

/* In some cases a fpdu will arrive with only one byte of the header, in this
 * case the fpdu_length will be partial ( contain only higher byte and
 * incomplete bytes will contain the invalid value */
#define QED_IWARP_INVALID_INCOMPLETE_BYTES 0xffff

struct qed_iwarp_fpdu {
	struct qed_iwarp_ll2_buff *mpa_buf;
	dma_addr_t pkt_hdr;
	u8 pkt_hdr_size;
	dma_addr_t mpa_frag;
	void *mpa_frag_virt;
	u16 mpa_frag_len;
	u16 fpdu_length;
	u16 incomplete_bytes;
};

struct qed_iwarp_info {
	struct list_head listen_list;	/* qed_iwarp_listener */
	struct list_head ep_list;	/* qed_iwarp_ep */
	struct list_head ep_free_list;	/* pre-allocated ep's */
	struct list_head mpa_buf_list;	/* list of mpa_bufs */
	struct list_head mpa_buf_pending_list;
	spinlock_t iw_lock;
	spinlock_t qp_lock;	/* for teardown races */
	struct iwarp_rxmit_stats_drv stats;
	u32 rcv_wnd_scale;
	u16 max_mtu;
	u16 num_ooo_rx_bufs;
	u8 mac_addr[ETH_ALEN];
	u8 crc_needed;
	u8 tcp_flags;
	u8 ll2_syn_handle;
	u8 ll2_ooo_handle;
	u8 ll2_mpa_handle;
	u8 peer2peer;
	u8 _pad;
	enum mpa_negotiation_mode mpa_rev;
	enum mpa_rtr_type rtr_type;
	struct qed_iwarp_fpdu *partial_fpdus;
	struct qed_iwarp_ll2_mpa_buf *mpa_bufs;
	u8 *mpa_intermediate_buf;
	u16 max_num_partial_fpdus;

	/* MPA statistics */
	u64 unalign_rx_comp;
};
#endif

#define IS_QED_DCQCN(p_hwfn) \
	(!!(p_hwfn->pf_params.rdma_pf_params.enable_dcqcn))

struct qed_roce_info {
	struct roce_events_stats event_stats;

	u8 dcqcn_enabled;
	u8 dcqcn_reaction_point;
};

struct qed_rdma_info {
	spinlock_t lock;

	struct qed_bmap cq_map;
	struct qed_bmap pd_map;
	struct qed_bmap tid_map;
	struct qed_bmap srq_map;
	struct qed_bmap cid_map;
	struct qed_bmap tcp_cid_map;
	struct qed_bmap real_cid_map;
	struct qed_bmap dpi_map;
	struct qed_bmap toggle_bits;
	struct qed_rdma_events events;
	struct qed_rdma_device *dev;
	struct qed_rdma_port *port;
	u32 last_tid;
	u8 num_cnqs;
	struct rdma_sent_stats rdma_sent_pstats;
	struct rdma_rcv_stats rdma_rcv_tstats;
	u32 num_qps;
	u32 num_mrs;
	u32 num_srqs;
	u16 queue_zone_base;
	u16 max_queue_zones;
	enum protocol_type proto;
	struct qed_roce_info roce;
#ifdef CONFIG_IWARP
	struct qed_iwarp_info iwarp;
#endif
};

#ifdef CONFIG_IWARP
enum qed_iwarp_qp_state {
	QED_IWARP_QP_STATE_IDLE,
	QED_IWARP_QP_STATE_RTS,
	QED_IWARP_QP_STATE_TERMINATE,
	QED_IWARP_QP_STATE_CLOSING,
	QED_IWARP_QP_STATE_ERROR,
};
#endif

struct qed_rdma_qp {
	struct regpair qp_handle;
	struct regpair qp_handle_async;
	u32 qpid;		/* iwarp: may differ from icid */
	u16 icid;
	enum qed_roce_qp_state cur_state;
#ifdef CONFIG_IWARP
	enum qed_iwarp_qp_state iwarp_state;
#endif
	bool use_srq;
	bool signal_all;
	bool fmr_and_reserved_lkey;

	bool incoming_rdma_read_en;
	bool incoming_rdma_write_en;
	bool incoming_atomic_en;
	bool e2e_flow_control_en;

	u16 pd;			/* Protection domain */
	u16 pkey;		/* Primary P_key index */
	u32 dest_qp;
	u16 mtu;
	u16 srq_id;
	u8 traffic_class_tos;	/* IPv6/GRH traffic class; IPv4 TOS */
	u8 hop_limit_ttl;	/* IPv6/GRH hop limit; IPv4 TTL */
	u16 dpi;
	u32 flow_label;		/* ignored in IPv4 */
	u16 vlan_id;
	u32 ack_timeout;
	u8 retry_cnt;
	u8 rnr_retry_cnt;
	u8 min_rnr_nak_timer;
	bool sqd_async;
	union qed_gid sgid;	/* GRH SGID; IPv4/6 Source IP */
	union qed_gid dgid;	/* GRH DGID; IPv4/6 Destination IP */
	enum roce_mode roce_mode;
	u16 udp_src_port;	/* RoCEv2 only */
	u8 stats_queue;

	/* requeseter */
	u8 max_rd_atomic_req;
	u32 sq_psn;
	u16 sq_cq_id;		/* The cq to be associated with the send queue */
	u16 sq_num_pages;
	dma_addr_t sq_pbl_ptr;
	void *orq;
	dma_addr_t orq_phys_addr;
	u8 orq_num_pages;
	bool req_offloaded;

	/* responder */
	u8 max_rd_atomic_resp;
	u32 rq_psn;
	u16 rq_cq_id;		/* The cq to be associated with the receive queue */
	u16 rq_num_pages;
	dma_addr_t rq_pbl_ptr;
	void *irq;
	dma_addr_t irq_phys_addr;
	u8 irq_num_pages;
	bool resp_offloaded;
	u32 cq_prod;

	u8 remote_mac_addr[6];
	u8 local_mac_addr[6];

	void *shared_queue;
	dma_addr_t shared_queue_phys_addr;
#ifdef CONFIG_IWARP
	struct qed_iwarp_ep *ep;
#endif
};

#ifdef CONFIG_IWARP

enum qed_iwarp_ep_state {
	QED_IWARP_EP_INIT,
	QED_IWARP_EP_MPA_REQ_RCVD,
	QED_IWARP_EP_ESTABLISHED,
	QED_IWARP_EP_CLOSED
};

union async_output {
	struct iwarp_eqe_data_mpa_async_completion mpa_response;
	struct iwarp_eqe_data_tcp_async_completion mpa_request;
};

/* Endpoint structure represents a TCP connection. This connection can be
 * associated with a QP or not (in which case QP==NULL)
 */
struct qed_iwarp_ep {
	struct list_head list_entry;
	int sig;
	struct qed_rdma_qp *qp;
	enum qed_iwarp_ep_state state;

	/* This contains entire buffer required for ep memories. This is the
	 * only one actually allocated and freed. The rest are pointers into
	 * this buffer
	 */
	void *ep_buffer_virt;
	dma_addr_t ep_buffer_phys;

	/* Asynce EQE events contain only the ep pointer on the completion. The
	 * rest of the data is written to an output buffer pre-allocated by
	 * the driver. This buffer points to a location in the ep_buffer.
	 */
	union async_output *async_output_virt;
	dma_addr_t async_output_phys;

	struct qed_iwarp_cm_info cm_info;
	enum tcp_connect_mode connect_mode;
	enum mpa_rtr_type rtr_type;
	enum mpa_negotiation_mode mpa_rev;
	u32 tcp_cid;
	u32 cid;
	u8 remote_mac_addr[6];
	u8 local_mac_addr[6];
	u16 mss;
	bool mpa_reply_processed;

	/* The event_cb function is called for asynchrounous events associated
	 * with the ep. It is initialized at different entry points depending
	 * on whether the ep is the tcp connection active side or passive side
	 * The cb_context is passed to the event_cb function.
	 */
	iwarp_event_handler event_cb;
	void *cb_context;

	/* For Passive side - syn packet related data */
	struct qed_iwarp_ll2_buff *syn;
	u16 syn_ip_payload_length;
	dma_addr_t syn_phy_addr;
};

struct qed_iwarp_listener {
	struct list_head list_entry;

	/* The event_cb function is called for connection requests.
	 * The cb_context is passed to the event_cb function.
	 */
	iwarp_event_handler event_cb;
	void *cb_context;
	u32 max_backlog;
	u8 ip_version;
	u32 ip_addr[4];
	u16 port;
	u16 vlan;
};

void qed_iwarp_async_event(struct qed_hwfn *p_hwfn,
			   u8 fw_event_code,
			   struct regpair *fw_handle, u8 fw_return_code);

#endif /* CONFIG_IWARP */

void qed_roce_async_event(struct qed_hwfn *p_hwfn,
			  u8 fw_event_code, union rdma_eqe_data *rdma_data);

#endif
